import {
  createBrowserHistory,
  History,
  Location,
  LocationDescriptorObject,
  Path,
} from 'history'
import assign from 'lodash/assign'
import flatten from 'lodash/flatten'
import { action, computed, observable } from 'mobx'

export const browserHistory: History<any> = createBrowserHistory()
const basePush = browserHistory.push.bind(browserHistory)
const baseReplace = browserHistory.replace.bind(browserHistory)
const baseGoBack = browserHistory.goBack.bind(browserHistory)

const internalKey = Symbol('internalForward')

const canGoBackInternally = () =>
  !!(
    browserHistory.location.state && browserHistory.location.state[internalKey]
  )

const markStateAsInternal = (state: any = {}) => ({
  ...state,
  [internalKey]: true,
})

const maintainInternalState = (state?: any) =>
  canGoBackInternally() ? markStateAsInternal(state) : state

const overrideHistory: Partial<History> = {
  push(location: Path | LocationDescriptorObject, state: any = {}) {
    if (typeof location === 'string') {
      basePush(location, markStateAsInternal(state))
    } else {
      basePush({
        ...location,
        state: markStateAsInternal(location.state),
      })
    }
  },
  replace(location: Path | LocationDescriptorObject, state: any = {}) {
    if (typeof location === 'string') {
      baseReplace(location, maintainInternalState(state))
    } else {
      baseReplace({
        ...location,
        state: maintainInternalState(location.state),
      })
    }
  },
  goBack() {
    if (!canGoBackInternally()) {
      baseReplace(
        locationStore.isAtTopRoute ? '/' : locationStore.currentTopRoutePath,
      )
      return
    }
    baseGoBack()
  },
}

assign(browserHistory, overrideHistory)

type Disposer = () => void

class LocationStore {
  @observable
  public path = browserHistory.location.pathname

  @computed
  get isAtTopRoute(): boolean {
    return this.topRoutePaths.includes(this.path)
  }

  @computed
  get currentTopRoutePath(): string {
    let bestAncestor: string = '/'
    this.topRoutePaths.forEach(p => {
      if (p.length <= bestAncestor.length) {
        return
      }
      if (p === this.path) {
        bestAncestor = p
      } else if (this.path.startsWith(`${p}/`)) {
        bestAncestor = p
      }
    })
    return bestAncestor
  }

  @observable
  private keyMapTopRoutePaths: Map<number, string[]> = new Map()

  private nextKey = 1

  @computed
  private get topRoutePaths(): string[] {
    const result = flatten([...this.keyMapTopRoutePaths.values()])
    if (!result.includes('/')) {
      result.push('/')
    }
    return result
  }

  constructor() {
    browserHistory.listen(
      action((location: Location) => {
        this.path = location.pathname
      }),
    )
  }

  public isCurrentTopRoute = (path: string): boolean => {
    return this.currentTopRoutePath === path
  }

  @action
  public addTopRoutePaths(topRoutePaths: string[]): Disposer {
    const key = this.nextKey
    this.nextKey += 1
    this.keyMapTopRoutePaths.set(key, topRoutePaths)
    return () => {
      this.keyMapTopRoutePaths.delete(key)
    }
  }
}

export const locationStore = new LocationStore()
